"""
A module for spatial weights
"""
from .queen import queen_weights
from .rook import rook_weights
from .distance import distance_weights, min_distthreshold, knn_weights
from .kernel import kernel_weights, kernel_knn_weights
from .weight import Weight, read_gal, read_gwt, read_swm


from .redcap import redcap
from .skater import skater
from .maxp import maxp_greedy, maxp_sa, maxp_tabu
from .azp import azp_greedy, azp_sa, azp_tabu
from .schc import schc
from .validation import spatial_validation, make_spatial, ValidationResult, Diameter, Compactness, JoinCountRatio, Fragmentation


__version__ = "0.0.8-1"
# __version__ has to be define in the first line

__author__ = "Xun Li <lixun910@gmail.com>, "

# exposed under pygeoda.xxx
from .weights import *
from .clustering import *
from .sa import *
from .gda import *
from .classify import *
from .data import *



from ..libgeoda import gda_localmoran, gda_bi_localmoran, VecBool, gda_batchlocalmoran, VecVecBool, gda_localmoran_eb
from .lisa import lisa, batchlisa

__author__ = "Xun Li <lixun910@gmail.com>"

def local_moran(w, data, **kwargs):
    """Local Moran statistics.

    Args:
        w (Weight): An instance of Weight class.
        data (tuple/list/pandas.Series): A tuple of numeric values of selected variable
        undefs (tuple/list, optional): A tuple of boolean values to indicate which value is undefined or null
        permutations (int, optional): The number of permutations for the LISA computation
        permutation_method (str, optional): The permutation method used for the LISA computation. Options are {'complete', 'lookup-table'}. Default is 'complete'.
        significance_cutoff (float, optional): A cutoff value for significance p-values to filter not-significant clusters
        cpu_threads (int, optional): The number of cpu threads used for parallel LISA computation
        seed (int, optional): The seed for random number generator

    Returns:
        lisa: An instance of lisa class represents the results of lisa computation
    """
    if w == None:
        raise ValueError("Weights is None.")

    if w.num_obs != len(data):
        raise ValueError("The size of data doesnt not match the number of observations.")

    undefs = VecBool() if 'undefs' not in kwargs else kwargs['undefs']
    significance_cutoff = 0.05 if 'significance_cutoff' not in kwargs else kwargs['significance_cutoff']
    permutations =  999 if 'permutations' not in kwargs else kwargs['permutations']
    permutation_method = 'complete' if 'permutation_method' not in kwargs else kwargs['permutation_method']
    cpu_threads =  6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']
    seed =  123456789 if 'seed' not in kwargs else kwargs['seed']

    lisa_obj = gda_localmoran(w.gda_w, list(data), list(undefs), significance_cutoff, cpu_threads, permutations, permutation_method, seed)
    return lisa(lisa_obj)

def local_bimoran(w, data1, data2, **kwargs):
    """Bivariate local Moran statistics.
    The bivariate Local Moran’s I captures the relationship between the value for one variable at location i, and the average of the neighboring values for another variable.
    
    Args:
        w (Weight): An instance of Weight class.
        data1 (tuple/list/pandas.Series): A tuple of numeric values of first variable
        data1 (tuple/list/pandas.Series): A tuple of numeric values of second variable
        undefs (tuple/list, optional): A tuple of boolean values to indicate which value is undefined or null
        permutations (int, optional): The number of permutations for the LISA computation
        permutation_method (str, optional): The permutation method used for the LISA computation. Options are {'complete', 'lookup-table'}. Default is 'complete'.
        significance_cutoff (float, optional): A cutoff value for significance p-values to filter not-significant clusters
        cpu_threads (int, optional): The number of cpu threads used for parallel LISA computation
        seed (int, optional): The seed for random number generator

    Returns:
        lisa: An instance of lisa class represents the results of lisa computation
    """
    if w == None:
        raise ValueError("Weights is None.")

    if w.num_obs != len(data1) or len(data1) != len(data2):
        raise ValueError("The size of data doesnt not match the number of observations.")

    undefs = VecBool() if 'undefs' not in kwargs else kwargs['undefs']
    significance_cutoff = 0.05 if 'significance_cutoff' not in kwargs else kwargs['significance_cutoff']
    permutations =  999 if 'permutations' not in kwargs else kwargs['permutations']
    permutation_method = 'complete' if 'permutation_method' not in kwargs else kwargs['permutation_method']
    cpu_threads =  6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']
    seed =  123456789 if 'seed' not in kwargs else kwargs['seed']

    lisa_obj = gda_bi_localmoran(w.gda_w, list(data1), list(data2), list(undefs), significance_cutoff, cpu_threads, permutations, permutation_method, seed)
    return lisa(lisa_obj)

def local_moran_eb(w, event_data, base_data, **kwargs):
    """The function to apply local Moran with EB Rate statistics. 
    The EB rate is first computed from "event" and "base" variables, and then used in local moran statistics.

    Args:
        w (Weight): An instance of Weight class
        event_data (tuple/list/pandas.Series): A numeric tuple of selected "event" variable
        base_data (tuple/list/pandas.Series): A numeric tuple of selected "base" variable
        permutations (int, optional): The number of permutations for the LISA computation
        permutation_method (str, optional): The permutation method used for the LISA computation. Options are {'complete', 'lookup-table'}. Default is 'complete'.
        significance_cutoff (float, optional): A cutoff value for significance p-values to filter not-significant clusters
        cpu_threads (int, optional): The number of cpu threads used for parallel LISA computation
        seed (int, optional): The seed for random number generator

    Returns:
        lisa: An instance of lisa class represents the results of lisa computation
    """
    if w == None:
        raise ValueError("Weights is None.")

    if w.num_obs != len(event_data) or w.num_obs != len(base_data):
        raise ValueError("The size of data doesnt not match the number of observations.")

    significance_cutoff = 0.05 if 'significance_cutoff' not in kwargs else kwargs['significance_cutoff']
    permutations =  999 if 'permutations' not in kwargs else kwargs['permutations']
    permutation_method = 'complete' if 'permutation_method' not in kwargs else kwargs['permutation_method']
    cpu_threads =  6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']
    seed =  123456789 if 'seed' not in kwargs else kwargs['seed']

    lisa_obj = gda_localmoran_eb(w.gda_w, list(event_data), list(base_data), significance_cutoff, cpu_threads, permutations, permutation_method, seed)
    return lisa(lisa_obj)

def batch_local_moran(w, data, **kwargs):
    """Apply local moran statistics on a set of variables

    Args:
        w (Weight): An instance of Weight class.
        data (list/pandas.dataframe): A list of numeric array of selected variables
        permutations (int, optional): The number of permutations for the LISA computation
        permutation_method (str, optional): The permutation method used for the LISA computation. Options are {'complete', 'lookup-table'}. Default is 'complete'.
        significance_cutoff (float, optional): A cutoff value for significance p-values to filter not-significant clusters
        cpu_threads (int, optional): The number of cpu threads used for parallel LISA computation
        seed (int, optional): The seed for random number generator

    Returns:
        lisa: An instance of BatchLisa class represents the results of lisa computations
    """
    if w == None:
        raise ValueError("Weights is None.")

    if (not isinstance(data, list)) or (len(data) < 1):
        raise ValueError("The data has to be a list of numeric array of selected variables.")

    if w.num_obs != len(data[0]):   
        raise ValueError("The size of data doesnt not match the number of observations.")

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    undefs = VecVecBool() if 'undefs' not in kwargs else kwargs['undefs']
    permutations =  999 if 'permutations' not in kwargs else kwargs['permutations']
    permutation_method = 'complete' if 'permutation_method' not in kwargs else kwargs['permutation_method']
    significance_cutoff = 0.05 if 'significance_cutoff' not in kwargs else kwargs['significance_cutoff']
    cpu_threads =  6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']
    seed =  123456789 if 'seed' not in kwargs else kwargs['seed']

    lisa_obj = gda_batchlocalmoran(w.gda_w, data, undefs, significance_cutoff, cpu_threads, permutations, permutation_method, seed)
    return batchlisa(lisa_obj)


__author__ = "Xun Li <lixun910@gmail.com>"

from ..libgeoda import VecBool, VecVecBool
from ..libgeoda import gda_neighbor_match_test, VecVecDouble
from .lisa import lisa
import math

'''
Changes:
1/21/2021 Update local_multigeary, local_geary for 0.0.6
'''

def neighbor_match_test(geoda_obj, data, k, **kwargs):
    '''Local Neighbor Match Test
    The local neighbor match test is to assess the extent of overlap between k-nearest neighbors in geographical space and k-nearest neighbors in multi-attribute space.

    Args:
        geoda_obj (geoda): An instance of geoda class. 
        data (list or dataframe):   A list of numeric vectors of selected variable or a data frame of selected variables e.g. guerry[['Crm_prs', 'Literacy']]
        k (int): A positive integer number for k-nearest neighbors searching.
        scale_method (str, optional): One of the scaling methods {'raw', 'standardize', 'demean', 'mad', 'range_standardize', 'range_adjust'} to apply on input data. Default is 'standardize' (Z-score normalization).
        distance_method (str, optional): The type of distance metrics used to measure the distance between input data. Options are {'euclidean', 'manhattan'}. Default is 'euclidean'.
        power (float, optional): The power (or exponent) of a number says how many times to use the number in a multiplication.
        is_inverse (bool, optional): FALSE (default) or TRUE, apply inverse on distance value.
        is_arc (bool, optional): FALSE (default) or TRUE, compute arc distance between two observations.
        is_mile (bool, optional) TRUE (default) or FALSE, convert distance unit from mile to km.

    Returns:
        dict: A dict with values for key "Cardinality" and "Probability".
    '''
    scale_method = 'standardize' if 'scale_method' not in kwargs else kwargs['scale_method']
    distance_method = 'euclidean' if 'distance_method' not in kwargs else kwargs['distance_method']
    power = 1.0 if 'power' not in kwargs else kwargs['power']
    is_inverse =  False if 'is_inverse' not in kwargs else kwargs['is_inverse']
    is_arc =  False if 'is_arc' not in kwargs else kwargs['is_arc']
    is_mile =  False if 'is_mile' not in kwargs else kwargs['is_mile']

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    result = gda_neighbor_match_test(geoda_obj.gda, k, power, is_inverse, is_arc, is_mile, data, scale_method, distance_method)
    card = result[0]
    prob = result[1]

    new_card = [int(i) for i in card]
    # check p value
    new_prob = [p if p >= 0 else math.nan for p in prob]

    return {
        "Cardinality": new_card,
        "Probability": new_prob
    }



from ..libgeoda import gda_hinge15breaks
from ..libgeoda import gda_hinge30breaks
from ..libgeoda import gda_naturalbreaks
from ..libgeoda import gda_quantilebreaks
from ..libgeoda import gda_percentilebreaks
from ..libgeoda import gda_stddevbreaks
from ..libgeoda import VecBool

__author__ = "Xun Li <lixun910@gmail.com>, Hang Zhang <zhanghanggis@163.com>, "

def hinge15_breaks(data,**kwargs):
    """(Box) Hinge15 Breaks
    
    Hinge15 breaks data into 6 groups like box plot groups (Lower outlier, < 25%, 25-50%, 50-75%, >75%, Upper outlier) with hinge=1.5

    Args:
        data (tuple): A tuple with values of a selected variable
    
    Returns:
        list: A numeric list of break values
    """
    if data is None:
        raise("The data from selected variable is empty.")

    undefs = VecBool() if 'undefs' not in kwargs else kwargs['undefs']

    return gda_hinge15breaks(data, undefs)

def hinge30_breaks(data,**kwargs):
    """(Box) Hinge30 Breaks

    Hinge30 breaks data into 6 groups like box plot groups (Lower outlier, < 25%, 25-50%, 50-75%, >75%, Upper outlier) with hinge=3.0

    Args:
        data (tuple): A tuple with values of a selected variable
    
    Returns:
        list: A numeric list of break values
    """
    if data is None:
        raise("The data from selected variable is empty.")

    undefs = VecBool() if 'undefs' not in kwargs else kwargs['undefs']

    return gda_hinge30breaks(data, undefs)

def natural_breaks(k,data,**kwargs):
    """Natural Breaks (Jenks)
    
    Natural Breaks group data whose boundaries are set where there are relatively big differences.

    Args:
        k (int): Number of breaks
        data (tuple): A tuple with values of a selected variable
    
    Returns:
        list: A numeric list of break values
    """
    if k is None or k < 2:
        raise("The value of K needs to be greater than or equal to 2 ")

    if data is None:
        raise("The data from selected variable is empty.")

    undefs = VecBool() if 'undefs' not in kwargs else kwargs['undefs']
    return gda_naturalbreaks(k, data, undefs)

def quantile_breaks(k,data,**kwargs):
    """Quantile breaks

    Quantile breaks data into groups that each have the same number of observations

    Args:
        k (int): Number of breaks
        data (tuple): A tuple with values of a selected variable
    
    Returns:
        list: A numeric list of break values
    """
    if k is None or k < 2:
        raise("The value of K needs to be greater than or equal to 2 ")

    if data is None:
        raise("The data from selected variable is empty.")

    undefs = VecBool() if 'undefs' not in kwargs else kwargs['undefs']
    return gda_quantilebreaks(k, data, undefs)

def percentile_breaks(data,**kwargs):
    """Percentile breaks

    Note:
        The percentile breaks divides the data to  six ranges, the lowest 1%, 1-10%, 10-50%, 50-90%, 90-99% and the top 1%, returning the range boundaries as a breakpoint list. 

    Args:
        data (tuple): A tuple with values of a selected variable
    
    Returns:
        list: A numeric list of break values
    """
    if data is None:
        raise("The data from selected variable is empty.")

    undefs = VecBool() if 'undefs' not in kwargs else kwargs['undefs']
    return gda_percentilebreaks(data, undefs)

def stddev_breaks(data,**kwargs):
    """Standard deviation breaks

    Note:
        Standard deviation breaks calculate the number of standard deviational units of  the range from lowest to highest, returning a breakpoint list. 
    Args:
        data (tuple): A tuple with values of a selected variable
    
    Returns:
        list: A numeric list of break values
    """
    if data is None:
        raise("The data from selected variable is empty.")

    undefs = VecBool() if 'undefs' not in kwargs else kwargs['undefs']

    return gda_stddevbreaks(data, undefs)




__author__ = "Xun Li <lixun910@gmail.com>"

from ..libgeoda import gda_distance_weights, gda_min_distthreshold, gda_knn_weights
from .weight import Weight

def distance_weights(geoda_obj, dist_thres, **kwargs):
    '''Distance-based Spatial Weights
    Create a distance-based weights

    Args:
        geoda_obj (geoda): An instance of geoda class. 
        dist_thres (float): A positive numeric value of distance threshold used to find neighbors.
            For example, one can use the pygeoda.min_distthreshold() to get a distance that 
            guarantees that every observation has at least 1 neighbor.
        power (float, optional): The power (or exponent) of a number indicates how many times to use the number in a multiplication.
        is_inverse (bool, optional):  A bool flag indicates whether or not to apply inverse on distance value.
            Defaults to False.
        is_arc (bool, optional): A bool flag indicates if compute arc distance or Euclidean distance. Defaults to False (Euclidean distance) 
        is_mile (bool, optional): A bool flag indicates if the distance unit is mile or km. Defaults to True (mile).

    Returns:
        Weight: An instance of Weight class 
    '''
    power = 1.0 if 'power' not in kwargs else kwargs['power']
    is_inverse = False if 'is_inverse' not in kwargs else kwargs['is_inverse']
    is_arc = False if 'is_arc' not in kwargs else kwargs['is_arc']
    is_mile = True if 'is_mile' not in kwargs else kwargs['is_mile']

    poly_id = ""
    kernel = ""
    diagonal = False
    gda_w = gda_distance_weights(geoda_obj.gda, dist_thres, poly_id, power, is_inverse, is_arc, is_mile, kernel, diagonal)

    return Weight(gda_w)


def min_distthreshold(geoda_obj, is_arc=False, is_mile=True):
    '''Minimum Distance Threshold for Distance-based Weights
    Get minimum threshold of distance that makes sure each observation has at least one neighbor   

    Args:
        geoda_obj (geoda): An instance of geoda class. 
        is_arc (bool, optional): A bool flag indicates if compute arc distance or Euclidean distance. Defaults to False (Euclidean distance) 
        is_mile (bool, optional): A bool flag indicates if the distance unit is mile or km. Defaults to True (mile).

    Returns:
        thres (float): A float value of minimum threhold for distance based weights.
    '''
    return gda_min_distthreshold(geoda_obj.gda, is_arc, is_mile)

def knn_weights(geoda_obj, k, **kwargs):
    '''K-Nearest Neighbors-based Spatial Weights
    Create a k-nearest neighbors based spatial weights

    Args:
        geoda_obj (geoda): An instance of geoda class. 
        k (int): A positive integer number for k-nearest neighbors
        power (float, optional): The power (or exponent) of a number indicates how many times to use the number in a multiplication.
        is_inverse (bool, optional):  A bool flag indicates whether or not to apply inverse on distance value.
            Defaults to False.
        is_arc (bool, optional): A bool flag indicates if compute arc distance or Euclidean distance. Defaults to False (Euclidean distance) 
        is_mile (bool, optional): A bool flag indicates if the distance unit is mile or km. Defaults to True (mile).


    Returns:
        Weight: An instance of Weight class 
    '''
    power = 1.0 if 'power' not in kwargs else kwargs['power']
    is_inverse = False if 'is_inverse' not in kwargs else kwargs['is_inverse']
    is_arc = False if 'is_arc' not in kwargs else kwargs['is_arc']
    is_mile = True if 'is_mile' not in kwargs else kwargs['is_mile']
    
    # not used
    kernel = "" 
    bandwidth = 0 
    adaptive_bandwidth = False 
    use_kernel_diagnals = False 
    polyid = ""

    gda_w = gda_knn_weights(geoda_obj.gda, k, power, is_inverse, is_arc, is_mile, kernel, bandwidth, adaptive_bandwidth, use_kernel_diagnals, polyid)

    return Weight(gda_w)


__author__ = "Xun Li <lixun910@gmail.com>"

from ..libgeoda import VecVecDouble, VecPair, VecDouble, VecInt, Pair
from ..libgeoda import gda_maxp_greedy, gda_maxp_sa, gda_maxp_tabu
from ..libgeoda import gda_betweensumofsquare, gda_totalsumofsquare, gda_withinsumofsquare, flat_2dclusters

'''
Changes:
1/20/2021 Add maxp_greedy, maxp_sa, maxp_tabu
'''

def maxp_greedy(w, data, bound_variable, min_bound, **kwargs):
    ''' A greedy algorithm to solve the max-p-region problem 
    
    The max-p-region problem is a special case of constrained clustering where a finite number of geographical areas are aggregated into the maximum number of regions (max-p-regions), such that each region is geographically connected and the clusters could maximize internal homogeneity.

    Arguments:
        w (Weight): an instance of Weight class
        data (list or dataframe):   A list of numeric vectors of selected variable or a data frame of selected variables e.g. guerry[['Crm_prs', 'Literacy']]
        bound_variable (tuple): A numeric vector of selected bounding variable
        min_bound (float): A minimum value that the sum value of bounding variable int each cluster should be greater than
        iterations (int, optional): The number of iterations of greedy algorithm. Defaults to 99.
        init_regions (tuple, optional): The initial regions that the local search starts with. Default is empty. means the local search starts with a random process to "grow" clusters
        scale_method (str, optional): One of the scaling methods {'raw', 'standardize', 'demean', 'mad', 'range_standardize', 'range_adjust'} to apply on input data. Default is 'standardize' (Z-score normalization).
        distance_method (str, optional):  The distance method used to compute the distance betwen observation i and j. Defaults to "euclidean". Options are "euclidean" and "manhattan"
        random_seed (int, optional): The seed for random number generator. Defaults to 123456789. It is the same as GeoDa software
        cpu_threads (int, optional): The number of cpu threads used for parallel computation

    Returns:
        dict: A dict with keys {"Clusters", "TotalSS", "Within-clusterSS", "TotalWithin-clusterSS", "Ratio"}
    '''
    
    iterations = 99 if 'iterations' not in kwargs else kwargs['iterations']
    init_regions = [] if 'init_regions' not in kwargs else kwargs['init_regions']
    scale_method = "standardize" if "scale_method" not in kwargs else kwargs['scale_method']
    distance_method = 'euclidean' if 'distance_method' not in kwargs else kwargs['distance_method'] 
    random_seed = 123456789 if 'random_seed' not in kwargs else kwargs['random_seed']
    cpu_threads = 6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']

    if w.num_obs < 1:
        raise ValueError("The weights is not valid.")

    if len(data) < 1:
        raise ValueError("The data from selected variable is empty.")

    if len(bound_variable) != w.num_obs:
        raise ValueError("The bound_variable has to be a list of numeric values, e.g. a column of input table.")

    if min_bound <= 0:
        raise ValueError("The min_bound has to be a positive numeric value.")
    
    in_data = VecVecDouble()

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    for d in data:
        in_data.push_back(d)

    min_bounds = VecPair()
    max_bounds = VecPair()

    min_bounds.push_back((min_bound, list(bound_variable)))

    in_init_regions = VecInt(list(init_regions))

    cluster_ids = gda_maxp_greedy(w.gda_w, in_data, scale_method, iterations, min_bounds, max_bounds, in_init_regions, distance_method, random_seed, cpu_threads)

    between_ss = gda_betweensumofsquare(cluster_ids, in_data)
    total_ss = gda_totalsumofsquare(in_data)
    ratio = between_ss / total_ss
    within_ss = gda_withinsumofsquare(cluster_ids, in_data)

    return {
        "Total sum of squares" : total_ss,
        "Within-cluster sum of squares" : within_ss,
        "Total within-cluster sum of squares" : between_ss,
        "The ratio of between to total sum of squares" : ratio,
        "Clusters" : flat_2dclusters(w.num_obs, cluster_ids),
    }

def maxp_sa(w, data, bound_variable, min_bound, cooling_rate=0.85, **kwargs):
    ''' A simulated annealing algorithm to solve the max-p-region problem 
    
    The max-p-region problem is a special case of constrained clustering where a finite number of geographical areas are aggregated into the maximum number of regions (max-p-regions), such that each region is geographically connected and the clusters could maximize internal homogeneity.
        
    Arguments:
        w (Weight): an instance of Weight class
        data (list or dataframe):   A list of numeric vectors of selected variable or a data frame of selected variables e.g. guerry[['Crm_prs', 'Literacy']]
        bound_variable (tuple): A numeric vector of selected bounding variable
        min_bound (float): A minimum value that the sum value of bounding variable int each cluster should be greater than
        cooling_rate (float): The cooling rate of a simulated annealing algorithm. Defaults to 0.85
        sa_maxit (int): The number of iterations of simulated annealing. Defaults to 1
        iterations (int, optional): The number of iterations of greedy algorithm. Defaults to 99.
        init_regions (tuple, optional): The initial regions that the local search starts with. Default is empty. means the local search starts with a random process to "grow" clusters
        scale_method (str, optional): One of the scaling methods {'raw', 'standardize', 'demean', 'mad', 'range_standardize', 'range_adjust'} to apply on input data. Default is 'standardize' (Z-score normalization).
        distance_method (str, optional):  The distance method used to compute the distance betwen observation i and j. Defaults to "euclidean". Options are "euclidean" and "manhattan"
        random_seed (int, optional): The seed for random number generator. Defaults to 123456789. It is the same as GeoDa software
        cpu_threads (int, optional): The number of cpu threads used for parallel computation

    Returns:
        dict: A dict with keys {"Clusters", "TotalSS", "Within-clusterSS", "TotalWithin-clusterSS", "Ratio"}
    '''

    sa_maxit = 1 if 'sa_maxit'  not in kwargs else kwargs['sa_maxit']
    iterations = 99 if 'iterations' not in kwargs else kwargs['iterations']
    init_regions = [] if 'init_regions' not in kwargs else kwargs['init_regions']
    scale_method = "standardize" if "scale_method" not in kwargs else kwargs['scale_method']
    distance_method = 'euclidean' if 'distance_method' not in kwargs else kwargs['distance_method'] 
    random_seed = 123456789 if 'random_seed' not in kwargs else kwargs['random_seed']
    cpu_threads = 6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']

    if cooling_rate <=0 or cooling_rate >=1:
        raise ValueError("Cooling rate should be in range (0,1).")

    if w.num_obs < 1:
        raise ValueError("The weights is not valid.")

    if len(data) < 1:
        raise ValueError("The data from selected variable is empty.")

    if len(bound_variable) != w.num_obs:
        raise ValueError("The bound_variable has to be a list of numeric values, e.g. a column of input table.")

    if min_bound <= 0:
        raise ValueError("The min_bound has to be a positive numeric value.")
    
    in_data = VecVecDouble()

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    for d in data:
        in_data.push_back(d)

    min_bounds = VecPair()
    max_bounds = VecPair()

    min_bounds.push_back((min_bound, list(bound_variable)))

    in_init_regions = VecInt(list(init_regions))

    cluster_ids = gda_maxp_sa(w.gda_w, in_data, scale_method, iterations, cooling_rate, sa_maxit, min_bounds, max_bounds, in_init_regions, distance_method, random_seed, cpu_threads)

    between_ss = gda_betweensumofsquare(cluster_ids, in_data)
    total_ss = gda_totalsumofsquare(in_data)
    ratio = between_ss / total_ss
    within_ss = gda_withinsumofsquare(cluster_ids, in_data)

    return {
       "Total sum of squares" : total_ss,
        "Within-cluster sum of squares" : within_ss,
        "Total within-cluster sum of squares" : between_ss,
        "The ratio of between to total sum of squares" : ratio,
        "Clusters" : flat_2dclusters(w.num_obs, cluster_ids),
    }

def maxp_tabu(w, data, bound_variable, min_bound, tabu_length=10, **kwargs):
    ''' A tabu-search algorithm to solve the max-p-region problem 
    
    The max-p-region problem is a special case of constrained clustering where a finite number of geographical areas are aggregated into the maximum number of regions (max-p-regions), such that each region is geographically connected and the clusters could maximize internal homogeneity.
        
    Arguments:
        w (Weight): an instance of Weight class
        data (list or dataframe):   A list of numeric vectors of selected variable or a data frame of selected variables e.g. guerry[['Crm_prs', 'Literacy']]
        bound_variable (tuple): A numeric vector of selected bounding variable
        min_bound (float): A minimum value that the sum value of bounding variable int each cluster should be greater than
        tabu_length (int): The length of a tabu search heuristic of tabu algorithm. Defaults to 10.
        conv_tabu (int, optional): The number of non-improving moves. Defaults to 10.
        iterations (int, optional): The number of iterations of greedy algorithm. Defaults to 99.
        init_regions (tuple, optional): The initial regions that the local search starts with. Default is empty. means the local search starts with a random process to "grow" clusters
        scale_method (str, optional): One of the scaling methods {'raw', 'standardize', 'demean', 'mad', 'range_standardize', 'range_adjust'} to apply on input data. Default is 'standardize' (Z-score normalization).
        distance_method (str, optional):  The distance method used to compute the distance betwen observation i and j. Defaults to "euclidean". Options are "euclidean" and "manhattan"
        random_seed (int, optional): The seed for random number generator. Defaults to 123456789. It is the same as GeoDa software
        cpu_threads (int, optional): The number of cpu threads used for parallel computation

    Returns:
        dict: A dict with keys {"Clusters", "TotalSS", "Within-clusterSS", "TotalWithin-clusterSS", "Ratio"}
    '''
    
    conv_tabu = 10 if 'conv_tabu' not in kwargs else kwargs['conv_tabu']
    iterations = 99 if 'iterations' not in kwargs else kwargs['iterations']
    init_regions = [] if 'init_regions' not in kwargs else kwargs['init_regions']
    scale_method = "standardize" if "scale_method" not in kwargs else kwargs['scale_method']
    distance_method = 'euclidean' if 'distance_method' not in kwargs else kwargs['distance_method'] 
    random_seed = 123456789 if 'random_seed' not in kwargs else kwargs['random_seed']
    cpu_threads = 6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']

    if w.num_obs < 1:
        raise ValueError("The weights is not valid.")

    if len(data) < 1:
        raise ValueError("The data from selected variable is empty.")

    if len(bound_variable) != w.num_obs:
        raise ValueError("The bound_variable has to be a list of numeric values, e.g. a column of input table.")

    if min_bound <= 0:
        raise ValueError("The min_bound has to be a positive numeric value.")
    
    in_data = VecVecDouble()

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    for d in data:
        in_data.push_back(d)

    min_bounds = VecPair()
    max_bounds = VecPair()

    min_bounds.push_back((min_bound, list(bound_variable)))

    in_init_regions = VecInt(list(init_regions))

    cluster_ids = gda_maxp_tabu(w.gda_w, in_data, scale_method, iterations, tabu_length, conv_tabu, min_bounds, max_bounds, in_init_regions, distance_method, random_seed, cpu_threads)

    between_ss = gda_betweensumofsquare(cluster_ids, in_data)
    total_ss = gda_totalsumofsquare(in_data)
    ratio = between_ss / total_ss
    within_ss = gda_withinsumofsquare(cluster_ids, in_data)

    return {
        "Total sum of squares" : total_ss,
        "Within-cluster sum of squares" : within_ss,
        "Total within-cluster sum of squares" : between_ss,
        "The ratio of between to total sum of squares" : ratio,
        "Clusters" : flat_2dclusters(w.num_obs, cluster_ids),
    }


from ..libgeoda import gda_rook_weights 
from .weight import Weight

__author__ = "Xun Li <lixun910@gmail.com>"

def rook_weights(geoda_obj, **kwargs):
    '''Rook Contiguity Spatial Weights
    Create a Rook contiguity weights with options of "order", "include lower order" and "precision threshold"
    
    Args:
        geoda_obj (geoda): An instance of geoda class
        order (int, optional): order of contiguity
        include_lower_order (bool, optional):  whether or not the lower order neighbors should be 
            included in the weights structure
        precision_threshold  (float, optional):  the precision of the underlying shape file is insufficient 
            to allow for an exact match of coordinates to determine 
            which polygons are neighbors

    Returns:
        Weight: An instance of Weight object
    '''
    order = 1 if 'order' not in kwargs else kwargs['order']
    include_lower_order = False if 'include_lower_order' not in kwargs else kwargs['include_lower_order']
    precision_threshold = 0.0 if 'precision_threshold' not in kwargs else kwargs['precision_threshold']
    
    gda_w = gda_rook_weights(geoda_obj.gda, order, include_lower_order, precision_threshold)

    return Weight(gda_w)




from ..libgeoda import gda_localjoincount, gda_localmultijoincount, VecBool, VecVecBool
from .lisa import lisa
import multiprocessing

__author__ = "Xun Li <lixun910@gmail.com>"

def local_joincount(w, data, **kwargs):
    """Local Join Count Statistics

    The function to apply local Join Count statistics.

    Args:
        w (Weight): An instance of Weight class.
        data (tuple/list/pandas.Series): A list of numeric values of selected variable
        undefs (list, optional): A list of boolean values to indicate which value is undefined or null
        permutations (int, optional): The number of permutations for the LISA computation
        permutation_method (str, optional): The permutation method used for the LISA computation. Options are {'complete', 'lookup-table'}. Default is 'complete'.
        significance_cutoff (float, optional): A cutoff value for significance p-values to filter not-significant clusters
        cpu_threads (int, optional): The number of cpu threads used for parallel LISA computation
        seed (int, optional): The seed for random number generator

    Returns:
        lisa: An instance of lisa class. 
    """
    if w == None:
        raise ValueError("Weights is None.")

    undefs = VecBool() if 'undefs' not in kwargs else kwargs['undefs']
    significance_cutoff = 0.05 if 'significance_cutoff' not in kwargs else kwargs['significance_cutoff']
    permutations =  999 if 'permutations' not in kwargs else kwargs['permutations']
    permutation_method = 'complete' if 'permutation_method' not in kwargs else kwargs['permutation_method']
    cpu_threads =  6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']
    seed =  123456789 if 'seed' not in kwargs else kwargs['seed']

    for i in data:
        if i != 0 and i != 1:
            raise ValueError("The input data is not binary.")

    lisa_obj = gda_localjoincount(w.gda_w, list(data), list(undefs), significance_cutoff, cpu_threads, permutations, permutation_method, seed)
    return lisa(lisa_obj)

def local_bijoincount(w, data, **kwargs):
    """Bivariate Local Join Count Statistics

    The function to apply (no-colocation) bivariate local Join Count statistics. The bivariate local join count only applies on two variables with no-colocation.

    Args:
        w (Weight): An instance of Weight class
        data (list or dataframe):   A list of numeric vectors of selected variable or a data frame of selected variables e.g. guerry[['Crm_prs', 'Literacy']]
        permutations (int, optional): The number of permutations for the LISA computation
        permutation_method (str, optional): The permutation method used for the LISA computation. Options are {'complete', 'lookup-table'}. Default is 'complete'.
        significance_cutoff (float, optional): A cutoff value for significance p-values to filter not-significant clusters
        cpu_threads (int, optional): The number of cpu threads used for parallel LISA computation
        seed (int, optional): The seed for random number generator

    Returns:
        lisa: An instance of lisa class 

    Examples:
        >>> import pygeoda
        >>> columbus = pygeoda.open("./data/columbus.shp")
        >>> columbus_q = pygeoda.queen_weights(columbus)
        >>> nsa = columbus.GetRealCol("nsa")
        >>> nsa_inv = [1-i for i in nsa]
        >>> lisa = pygeoda.local_bijoincount(columbus_q, [nsa, nsa_inv])
        >>> jc = lisa.lisa_values()
        (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 0.0, 3.0, 0.0, 3.0, 3.0, 2.0, 0.0, 0.0, 0.0, 0.0, 3.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
        >>> pvals = lisa.lisa_pvalues()
        (nan, nan, nan, nan, nan, nan, nan, nan, 0.002, 0.034, nan, nan, nan, nan, nan, nan, 0.44, nan, nan, nan, nan, 0.262, nan, 0.125, 0.079, 0.053, nan, nan, nan, nan, 0.093, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan, nan)
        >>> nn = lisa.lisa_num_nbrs()
        (2, 3, 4, 4, 8, 2, 4, 6, 8, 4, 5, 6, 4, 6, 6, 8, 3, 4, 3, 10, 3, 6, 3, 7, 8, 6, 4, 9, 7, 5, 3, 4, 4, 4, 7, 5, 6, 6, 3, 5, 3, 2, 6, 5, 4, 2, 2, 4, 3)
    """
    if w == None:
        raise ValueError("Weights is None.")

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    if len(data) != 2:
        raise ValueError("The input data should be a list of tuples of two selected variables.")
    
    undefs = VecVecBool() if 'undefs' not in kwargs else kwargs['undefs']
    significance_cutoff = 0.05 if 'significance_cutoff' not in kwargs else kwargs['significance_cutoff']
    permutations =  999 if 'permutations' not in kwargs else kwargs['permutations']
    permutation_method = 'complete' if 'permutation_method' not in kwargs else kwargs['permutation_method']
    cpu_threads =  6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']
    seed =  123456789 if 'seed' not in kwargs else kwargs['seed']

    data1 = data[0]
    data2 = data[1]

    n1 = len(data1)
    n2 = len(data2)

    if n1 != n2:
        raise ValueError("The size of data1 does not match the size of data2.")

    for d in data:
        for i in d:
            if i != 0 and i != 1:
                raise ValueError("The input data is not binary.")

    has_colocation = False
    for i in range(n1):
        if data1[i] == 1 and data2[i] == 1:
            has_colocation = True
            break

    if has_colocation:
        raise ValueError("The bivariate local join count only applies on two variables with no-colocation.")

    lisa_obj = gda_localmultijoincount(w.gda_w, data, undefs, significance_cutoff, cpu_threads, permutations, permutation_method, seed)
    return lisa(lisa_obj)

def local_multijoincount(w, data, **kwargs):
    """(Multivariate) Colocation Local Join Count Statistics

    The function to apply (multivariate) colocation local Join Count statistics

    Args:
        w (Weight): An instance of Weight class
        data (list or dataframe):   A list of numeric vectors of selected variable or a data frame of selected variables e.g. guerry[['Crm_prs', 'Literacy']]
        permutations (int, optional): The number of permutations for the LISA computation
        permutation_method (str, optional): The permutation method used for the LISA computation. Options are {'complete', 'lookup-table'}. Default is 'complete'.
        significance_cutoff (float, optional): A cutoff value for significance p-values to filter not-significant clusters
        cpu_threads (int, optional): The number of cpu threads used for parallel LISA computation
        seed (int, optional): The seed for random number generator

    Returns:
        lisa: An instance of lisa class 
    """
    if w == None:
        raise ValueError("Weights is None.")

    undefs = VecVecBool() if 'undefs' not in kwargs else kwargs['undefs']
    significance_cutoff = 0.05 if 'significance_cutoff' not in kwargs else kwargs['significance_cutoff']
    permutations =  999 if 'permutations' not in kwargs else kwargs['permutations']
    permutation_method = 'complete' if 'permutation_method' not in kwargs else kwargs['permutation_method']
    cpu_threads =  6 if 'cpu_threads' not in kwargs else kwargs['cpu_threads']
    seed =  123456789 if 'seed' not in kwargs else kwargs['seed']

    if type(data).__name__ == "DataFrame":
        data = data.values.transpose().tolist()

    n_vars = len(data)
    if n_vars <= 1:
        raise ValueError("The multivairate local Join Count applies to multiple (more than one) variables.")

    n_obs = len(data[0])
    for d in data:
        if len(d) != n_obs:
            raise ValueError("The input data have variables with different size.")
        for i in d:
            if i != 0 and i != 1:
                raise ValueError("The input data is not binary.")

    if n_vars == 2 and (sum(data[0]) + sum(data[1]) == n_obs):
        no_colocation = True
        for i in range(n_obs):
            if data[0][i] == 1 and data[1][i] == 1:
                no_colocation = False
                break
        if no_colocation: 
            raise ValueError("The input two variables have no colocations. Please use bivariate local join count: local_bijoincount().")

    lisa_obj = gda_localmultijoincount(w.gda_w, data, undefs, significance_cutoff, cpu_threads, permutations, permutation_method, seed)
    return lisa(lisa_obj)


__author__ = "Xun Li <lixun910@gmail.com>"

from ..libgeoda import gda_distance_weights, gda_knn_weights
from .weight import Weight

def kernel_weights(geoda_obj, bandwidth, kernel, **kwargs):
    '''Distance-based Kernel Spatial Weights
    Create a kernel weights by specifying a bandwidth and a kernel method

    Args:
        geoda_obj (geoda): An instance of geoda class. 
        bandwidth (float): A float value represents the distance of bandwidth used in kernel function.
            For example, one can use the pygeoda.min_distthreshold() to get a distance that 
            guarantees that every observation has at least 1 neighbor.
        kernel (str): A string value, which has to be one of {'triangular', 'uniform', 'epanechnikov', 'quartic', 'gaussian'}
        use_kernel_diagonals (bool, optional):  A bool flag indicates whether or not the lower order neighbors should be 
            included in the weights structure. Defaults to False.
        is_inverse (bool, optional):  A bool flag indicates whether or not to apply inverse on distance value. Defaults to False.
        power (float, optional): The power (or exponent) of a number indicates how many times to use the number in a multiplication.
        is_arc (bool, optional): A bool flag indicates if compute arc distance or Euclidean distance. Defaults to False (Euclidean distance) 
        is_mile (bool, optional): A bool flag indicates if the distance unit is mile or km. Defaults to True (mile).

    Returns:
        Weight: An instance of Weight class 
    '''
    if kernel not in ['triangular', 'uniform', 'epanechnikov', 'quartic', 'gaussian']:
        raise("The parameter 'kernel' has to be one of {'triangular', 'uniform', 'epanechnikov', 'quartic', 'gaussian'}")
    
    use_kernel_diagonals = False if 'use_kernel_diagonals' not in kwargs else kwargs['use_kernel_diagonals']
    is_arc = False if 'is_arc' not in kwargs else kwargs['is_arc']
    is_mile = True if 'is_mile' not in kwargs else kwargs['is_mile']
    power = 1.0 if 'power' not in kwargs else kwargs['power']
    is_inverse = False if 'is_inverse' not in kwargs else kwargs['is_inverse']

    poly_id = ""
    
    gda_w = gda_distance_weights(geoda_obj.gda, bandwidth, poly_id, power, is_inverse, is_arc, is_mile, kernel, use_kernel_diagonals)

    return Weight(gda_w)


def kernel_knn_weights(geoda_obj, k, kernel, **kwargs):
    '''K-NN Kernel Spatial Weights
    Create a kernel weights by specifying k-nearest neighbors and a kernel method

    Args:
        geoda_obj (geoda): An instance of geoda class. 
        k (int): a positive integer number for k-nearest neighbors
        kernel (str): A string value, which has to be one of {'triangular', 'uniform', 'epanechnikov', 'quartic', 'gaussian'}
        Adaptive_bandwidth (bool, optional): True(default) or False: True - use adaptive bandwidth calculated using distance of k-nearest neithbors, False - use max distance of all observation to their k-nearest neighbors
        use_kernel_diagonals (bool, optional):  A bool flag indicates whether or not the lower order neighbors should be 
            included in the weights structure. Defaults to False.
        is_inverse (bool, optional):  A bool flag indicates whether or not to apply inverse on distance value. Defaults to False.
        power (float, optional): The power (or exponent) of a number indicates how many times to use the number in a multiplication.
        is_arc (bool, optional): A bool flag indicates if compute arc distance or Euclidean distance. Defaults to False (Euclidean distance) 
        is_mile (bool, optional): A bool flag indicates if the distance unit is mile or km. Defaults to True (mile).

    Returns:
        Weight: An instance of Weight class 
    '''
    if kernel not in ['triangular', 'uniform', 'epanechnikov', 'quartic', 'gaussian']:
        raise("The parameter 'kernel' has to be one of 'triangular', 'uniform', 'epanechnikov', 'quartic', 'gaussian'")

    adaptive_bandwidth = True if 'adaptive_bandwidth' not in kwargs else kwargs['adaptive_bandwidth']
    use_kernel_diagonals = False if 'use_kernel_diagonals' not in kwargs else kwargs['use_kernel_diagonals']
    power = 1.0 if 'power' not in kwargs else kwargs['power']
    is_inverse = False if 'is_inverse' not in kwargs else kwargs['is_inverse']
    is_arc = False if 'is_arc' not in kwargs else kwargs['is_arc']
    is_mile = True if 'is_mile' not in kwargs else kwargs['is_mile']
    bandwidth = 0
    polyid = ""
    
    gda_w = gda_knn_weights(geoda_obj.gda, k, power, is_inverse, is_arc, is_mile, kernel, bandwidth, adaptive_bandwidth, use_kernel_diagonals, polyid)

    return Weight(gda_w)


__author__ = "Xun Li <lixun910@gmail.com>"

from ..libgeoda import gda_spatialvalidation, gda_all_joincount_ratio, gda_makespatial

class Diameter:
    """Diameter Measure

    Attributes:
        steps (int): the longest shortest distance between any pairs
        ratio (float): the ratio of steps over the number of elements in the cluster
    """
    def __init__(self, diam):
        """Constructor of Diameter

        Args:
            diam (Object): An object/pointer of libgeoda Diameter class
        """
        self.steps = diam.steps
        self.ratio = diam.ratio

    def __repr__(self):
        info = "{"
        info += "'Steps': {0}, ".format(self.steps)
        info += "'Ratio': {0}".format(self.ratio)
        info += "}"
        return info

class Compactness:
    """Compactness Measure

    Attributes:
        area (float): the area of a cluster. If points, the convex hull is used to compute
            the area.
        perimeter (float): the perimeter of a cluster. If points, the convex hull is used
            to compute the perimeter
        isoperimeter_quotient (float): (4 * pi * area) / (perimeter^2)
    """
    def __init__(self, comp):
        """Constructor of Compactness

        Args:
            comp (Object): An object/pointer of libgeoda Compactness class
        """
        self.area = comp.area
        self.perimeter = comp.perimeter
        self.isoperimeter_quotient = comp.isoperimeter_quotient

    def __repr__(self):
        info = "{"
        info += "'Area': {0}, ".format(self.area)
        info += "'Perimeter': {0}, ".format(self.perimeter)
        info += "'Isoperimeter quotient': {0}".format(self.isoperimeter_quotient)
        info += "}"
        return info

class JoinCountRatio:
    """Join Count Ratio

    Join count ratio is the join counts, the number of times a category is surrounded 
    by neighbors of the same category, over the total number of neighbors after converting
    each category to a dummy variable.

    Attributes:
        n (int): the number of elements in a cluster
        neighbors (int): the total number of neighbors of elements in a cluster
        join_count (int): the total join count of elements in a cluster
        ratio (float): the ratio of total join count over total neighbors
    """
    def __init__(self, jcr):
        """Constructor of JoinCountRatio

        Args:
            jcr (Object): An object/pointer of libgeoda JoinCountRatio class
        """
        self.n = jcr.n
        self.neighbors = jcr.totalNeighbors
        self.join_count = jcr.totalJoinCount
        self.ratio = jcr.ratio

    def __repr__(self):
        info = "{"
        info += "'n': {0}, ".format(self.n)
        info += "'Neighbors': {0}, ".format(self.neighbors)
        info += "'Join Count': {0}, ".format(self.join_count)
        info += "'Ratio': {0}".format(self.ratio)
        info += "}"
        return info

class Fragmentation:
    """
    Fragmentation measure of spatial validation

    Attributes: 
        n (int): number of elements in clusters
        entropy (float): entropy measure for fraction of observations in each cluster
        std_entropy (float): standardized entropy measure
        simpson (float): simpson's index is a measure of diversity in each cluster
        std_simpson (float): standardized simpson measure
    """
    def __init__(self, gda_frag):
        """
        Constructor of Fragmentation class.

        Args:
            gda_frag (Object): A libgeoda Fragmentation instance
        """
        self.n = gda_frag.n
        self.entropy = gda_frag.entropy
        self.std_entropy = gda_frag.std_entropy
        self.simpson = gda_frag.simpson
        self.std_simpson = gda_frag.std_simpson

    def __repr__(self):
        info = "{"
        info += "'Entropy': {0}, ".format(self.entropy)
        info += "'Entropy*': {0}, ".format(self.std_entropy)
        info += "'Simpson': {0}, ".format(self.simpson)
        info += "'Simpson*': {0}".format(self.std_simpson)
        info += "}"
        return info

class ValidationResult:
    """
    Spatial Validation Result

    Attributes:
        spatially_constrained (bool): indicate if the clusters are 
        n (int): number of clusters
        joincount_ratio (JoinCountRatio): a list of JoinCountRatio objects
        fragmantation (Fragmentation): an object of Fragmentation
        cluster_fragmentation (Fragmentation): a list of Fragmentation objects, or None for spatially constrained clusters
        compactness (Compactness): a list of Compactness objects, or None for non-spatially constrained clusters
        diameter (Diameter): a list of Diameter objects, or None for non-spatially constrained clusters
    """
    def __init__(self, gda_validation):
        """
        Constructor of ValidationResult class.

        Args:
            gda_validation (Object): A libgeoda validation instance
        """
        self.spatially_constrained = gda_validation.spatially_constrained
        
        self.fragmentation = Fragmentation(gda_validation.fragmentation)
        self.n = self.fragmentation.n
        self.joincount_ratio = []
        
        for i in range(self.n):
            self.joincount_ratio.append(JoinCountRatio(gda_validation.joincount_ratio[i]))

        self.all_joincount_ratio = JoinCountRatio(gda_all_joincount_ratio(gda_validation.joincount_ratio))

        if self.spatially_constrained:
            self.cluster_fragmentation = None
            self.compactness = []
            self.diameter = []
            for i in range(self.n):
                self.compactness.append(Compactness(gda_validation.cluster_compactness[i]))
                self.diameter.append(Diameter(gda_validation.cluster_diameter[i]))
        else:
            self.cluster_fragmentation = []
            for i in range(self.n):
                self.cluster_fragmentation.append(Fragmentation(gda_validation.cluster_fragmentation[i]))
            self.compactness = None
            self.diameter = None

    def __repr__(self):
        info = "{"
        info += "'Spatially Constrained': {0}\n".format(self.spatially_constrained)
        info += "'Join Count Ratio': {0}\n".format(self.joincount_ratio)
        info += "'All Join Count Ratio': {0}\n".format(self.all_joincount_ratio)
        info += "'Fragmentation': {0}\n".format(self.fragmentation)
        info += "'Cluster Fragmentation': {0}\n".format(self.cluster_fragmentation)
        info += "'Compactness': {0}\n".format(self.compactness)
        info += "'Diameter': {0}\n".format(self.diameter)
        info += "}"
        return info

def spatial_validation(geoda_obj, clusters, w):
    """Spatial Validation
    Spatial validation provides a collection of validation measures including
    (1) fragmentations (entropy, simpson), (2) join count ratio, (3) compactness
    - isoperimeter quotient and (4) diameter.

    Args:
        geoda_obj (geoda): An instance of Geoda object
        clusters (list): A cluster classification variable (categorical values
        from a dataframe or values returned from cluster functions)
        w (Weight): An instance of Weight class.

    Returns:
        ValidationResult: An instance of ValidationResult including fragmentation, join count ratio,
                compactness, and diameter
    """
    if w is None:
        raise ValueError("Weights is None.")

    if w.num_obs != len(clusters):
        raise ValueError("The size of data doesnt not match the number of observations.")

    result = gda_spatialvalidation(geoda_obj.gda, clusters, w.gda_w)

    return ValidationResult(result)

def make_spatial(clusters, w):
    """Make spatially constrained clusters
    Make spatially constrained clusters from non-spatially constrained clusters

    Args:
        clusters (list): A cluster classification variable (categorical values
        from a dataframe or values returned from cluster functions, e.g. kmeans)
        w (Weight): An instance of Weight class.

    Returns:
        result (list): A new categorical values represent spatially constrained clusters
    """
    if w is None:
        raise ValueError("Weights is None.")

    if w.num_obs != len(clusters):
        raise ValueError("The size of data doesnt not match the number of observations.")

    result = gda_makespatial(list(clusters), w.gda_w)

    return result